@@ -1,9 +1,9 @@
 /*
  *  aespipe.c
  *
- *  Written by Jari Ruusu, February 11 2004
+ *  Written by Jari Ruusu, March 11 2005
  *
- *  Copyright 2002,2003,2004 by Jari Ruusu.
+ *  Copyright 2002-2005 by Jari Ruusu.
  *  Redistribution of this file is permitted under the GNU Public License.
  *
  *  AES encrypting or decrypting "pipe", reads from stdin, writes to stdout
@@ -69,9 +69,10 @@ union {
     u_int32_t       w[2048];
     unsigned char   b[8192];
 } buf;
-int             multiKeyMode = 0;
-char            *multiKeyPass[64];
+int             multiKeyMode = 0; /* 0=single-key 64=multi-key-v2 65=multi-key-v3 */
+char            *multiKeyPass[66];
 aes_context     multiKeyCtx[64];
+u_int32_t       partialMD5[4];
 
 int rd_wr_retry(int fd, char *buf, int cnt, int w)
 {
@@ -99,13 +100,24 @@ int rd_wr_retry(int fd, char *buf, int cnt, int w)
 
 char *get_FD_pass(int fd)
 {
-    char *p = NULL;
+    char *p = NULL, *n;
     int x = 0, y = 0;
 
     do {
         if(y >= (x - 1)) {
             x += 128;
-            if(!(p = realloc(p, x))) break;
+            /* Must enforce some max limit here.      */
+            /* This code may have successfully called */
+            /* mlockall(MCL_CURRENT | MCL_FUTURE)     */
+            if(x > (4*1024)) return(NULL);
+            n = malloc(x);
+            if(!n) return(NULL);
+            if(p) {
+                memcpy(n, p, y);
+                memset(p, 0, y);
+                free(p);
+            }
+            p = n;
         }
         if(rd_wr_retry(fd, p + y, 1, 0) != 1) break;
         if((p[y] == '\n') || !p[y]) break;
@@ -115,6 +127,13 @@ char *get_FD_pass(int fd)
     return p;
 }
 
+static void warnAboutBadKeyData(int x)
+{
+    if((x > 1) && (x != 64) && (x != 65)) {
+        fprintf(stderr, "Warning: Unknown key data format - using it anyway\n");
+    }
+}
+        
 char *do_GPG_pipe(char *pass)
 {
     int     x, pfdi[2], pfdo[2];
@@ -178,8 +197,7 @@ char *do_GPG_pipe(char *pass)
             a[x++] = "--homedir";
             a[x++] = gpgHomeDir;
         }
-        a[x++] = "--options";
-        a[x++] = "/dev/null";
+        a[x++] = "--no-options";
         a[x++] = "--quiet";
         a[x++] = "--batch";
         a[x++] = "--no-tty";
@@ -224,18 +242,21 @@ char *do_GPG_pipe(char *pass)
     close(pfdi[1]);
     memset(pass, 0, x);
     x = 0;
-    while(x < 64) {
+    while(x < 66) {
         multiKeyPass[x] = get_FD_pass(pfdo[0]);
         if(!multiKeyPass[x]) {
-            /* realloc() failed - abort */
+            /* mem alloc failed - abort */
             multiKeyPass[0] = 0;
             break;
         }
         if(strlen(multiKeyPass[x]) < AESPIPE_PASSWORD_MIN_LENGTH) break;
         x++;
     }
+    warnAboutBadKeyData(x);
+    if(x >= 65)
+        multiKeyMode = 65;
     if(x == 64)
-        multiKeyMode = 1;
+        multiKeyMode = 64;
     close(pfdo[0]);
     waitpid(gpid, &x, 0);
     if(!multiKeyPass[0]) goto nomem1;
@@ -275,7 +296,27 @@ char *sGetPass(int minLen)
         p = getPass("Password: ");
         ask2 = passAskTwice;
     } else {
-        p = get_FD_pass(passFDnumber);
+        if(gpgKeyFile && gpgKeyFile[0]) {
+            p = get_FD_pass(passFDnumber);
+        } else {
+            int x = 0;
+            while(x < 66) {
+                multiKeyPass[x] = get_FD_pass(passFDnumber);
+                if(!multiKeyPass[x]) goto nomem;
+                if(strlen(multiKeyPass[x]) < AESPIPE_PASSWORD_MIN_LENGTH) break;
+                x++;
+            }
+            warnAboutBadKeyData(x);
+            if(x >= 65) {
+                multiKeyMode = 65;
+                return multiKeyPass[0];
+            }
+            if(x == 64) {
+                multiKeyMode = 64;
+                return multiKeyPass[0];
+            }
+            p = multiKeyPass[0];
+        }
         ask2 = 0;
     }
     if(!p) goto nomem;
@@ -347,16 +388,11 @@ void rmd160HashTwiceWithA(unsigned char *ib, int ile, unsigned char *ob, int ole
     memset(pwdCopy, 0, sizeof(pwdCopy));
 }
 
-void compute_md5_iv(u_int32_t *data)
+void compute_md5_iv_v3(u_int32_t *data)
 {
     int         x, y, e;
     u_int32_t   buf[16];
 
-    iv[0] = 0x67452301;
-    iv[1] = 0xefcdab89;
-    iv[2] = 0x98badcfe;
-    iv[3] = 0x10325476;
-
     y = 7;
     e = 16;
     do {
@@ -368,6 +404,8 @@ void compute_md5_iv(u_int32_t *data)
             buf[12] = devSect0;
             buf[13] = (devSect1 & 0xFFFFFF) | 0x80000000;
             /* 4024 bits == 31 * 128 bit plaintext blocks + 56 bits of sector number */
+            /* For version 3 on-disk format this really should be 4536 bits, but can't be */
+            /* changed without breaking compatibility. V3 uses MD5-with-wrong-length IV */
             buf[14] = 4024;
             buf[15] = 0;
         }
@@ -399,7 +437,8 @@ void doEncryptMD5ivCBC(int cnt)
 
     do {
         a = &multiKeyCtx[devSect0 & 0x3F];
-        compute_md5_iv((u_int32_t *)(datap+16));
+        memcpy(&iv[0], &partialMD5[0], 16);
+        compute_md5_iv_v3((u_int32_t *)(datap+16));
         x = 32;
         do {
             iv[0] ^= *((u_int32_t *)(&datap[ 0]));
@@ -436,7 +475,8 @@ void doDecryptMD5ivCBC(int cnt)
             datap += 16;
         } while(--x);
         datap -= 512;
-        compute_md5_iv((u_int32_t *)(datap+16));
+        memcpy(&iv[0], &partialMD5[0], 16);
+        compute_md5_iv_v3((u_int32_t *)(datap+16));
         aes_decrypt(a, datap, datap);
         *((u_int32_t *)(&datap[ 0])) ^= iv[0];
         *((u_int32_t *)(&datap[ 4])) ^= iv[1];
@@ -500,7 +540,10 @@ int main(int argc, char **argv)
 {
     int x, encrypt = 1, bits, ret;
     void (*hf)(unsigned char *, int, unsigned char *, int);
-    unsigned char hb[32];
+    union {
+        u_int32_t     hw[16];
+        unsigned char hb[64];
+    } hbu;
     char *pass, *hfn = (char *)0, *efn = (char *)0;
     unsigned int y;
 
@@ -517,7 +560,7 @@ int main(int argc, char **argv)
         if(!strcmp(*argv, "-") || (**argv != '-')) {
             usage:
             fprintf(stderr, "usage: %s [options] <inputfile >outputfile\n"
-                            "version 2.2b  Copyright (c) 2002-2004 Jari Ruusu, (c) 2001 Dr Brian Gladman\n"
+                            "version 2.3b  Copyright (c) 2002-2005 Jari Ruusu, (c) 2001 Dr Brian Gladman\n"
                             "options:  -e aes128|aes192|aes256          =  set key length\n"
                             "          -H sha256|sha384|sha512|rmd160   =  set password hash function\n"
                             "          -d         =  decrypt\n"
@@ -614,55 +657,78 @@ int main(int argc, char **argv)
     pass = sGetPass((hf == rmd160HashTwiceWithA) ? 1 : AESPIPE_PASSWORD_MIN_LENGTH);
     if(!pass) exit(1);
     x = strlen(pass);
-    (*hf)((unsigned char *)pass, x, hb, sizeof(hb));
+    (*hf)((unsigned char *)pass, x, &hbu.hb[0], 32);
     if(multiKeyMode) {
         int r = 0, t;
-        while(r < 64) {
+        partialMD5[0] = 0x67452301;
+        partialMD5[1] = 0xefcdab89;
+        partialMD5[2] = 0x98badcfe;
+        partialMD5[3] = 0x10325476;
+        while(r < multiKeyMode) {
             t = strlen(multiKeyPass[r]);
-            (*hf)(multiKeyPass[r], t, hb, sizeof(hb));
+            (*hf)(multiKeyPass[r], t, &hbu.hb[0], 32);
             memset(multiKeyPass[r], 0, t);
             /*
              * MultiKeyMode uses md5 IV. One key mode uses sector IV. Sector IV
-             * and md5 IV are computed differently. This first key byte XOR with
-             * 0x55 is needed to cause complete decrypt failure in cases where
-             * data is encrypted with sector IV and decrypted with md5 IV or
-             * vice versa. If identical key was used but only IV was computed
-             * differently, only first plaintext block of 512 byte CBC chain
-             * would decrypt incorrectly and rest would decrypt correctly.
+             * and md5 IV v2 and v3 are all computed differently. This first key
+             * byte XOR with 0x55/0xF4 is needed to cause complete decrypt failure
+             * in cases where data is encrypted with one type of IV and decrypted
+             * with another type IV. If identical key was used but only IV was
+             * computed differently, only first plaintext block of 512 byte CBC
+             * chain would decrypt incorrectly and rest would decrypt correctly.
              * Partially correct decryption is dangerous. Decrypting all blocks
              * incorrectly is safer because file system mount will simply fail.
              */
-            hb[0] ^= 0x55;
-            aes_set_key(&multiKeyCtx[r], hb, bits, 0);
+            if(multiKeyMode == 65) {
+                hbu.hb[0] ^= 0xF4; /* version 3 */
+            } else {
+                hbu.hb[0] ^= 0x55; /* version 2 */
+            }
+            if(r < 64) {
+                aes_set_key(&multiKeyCtx[r], &hbu.hb[0], bits, 0);
+            } else {
+                /* only first 128 bits of iv-key is used */
+#if WORDS_BIGENDIAN
+                hbu.hw[0] = xcpu_to_le32(hbu.hw[0]);
+                hbu.hw[1] = xcpu_to_le32(hbu.hw[1]);
+                hbu.hw[2] = xcpu_to_le32(hbu.hw[2]);
+                hbu.hw[3] = xcpu_to_le32(hbu.hw[3]);
+#endif
+                memset(&hbu.hb[16], 0, 48);
+                md5_transform_CPUbyteorder(&partialMD5[0], &hbu.hw[0]);
+            }
             r++;
         }
     } else if(passIterThousands) {
         unsigned long iter = 0;
-        unsigned char tempkey[32];
+        union {
+            u_int32_t     w[8]; /* needed for 4 byte alignment of tempkey[] */
+            unsigned char tempkey[32];
+        } tku;
         /*
          * Set up AES-256 encryption key using same password and hash function
          * as before but with password bit 0 flipped before hashing. That key
          * is then used to encrypt actual encryption key N thousand times.
          */
         pass[0] ^= 1;
-        (*hf)((unsigned char *)pass, x, &tempkey[0], 32);
-        aes_set_key(&ctx, &tempkey[0], 256, 0);
+        (*hf)((unsigned char *)pass, x, &tku.tempkey[0], 32);
+        aes_set_key(&ctx, &tku.tempkey[0], 256, 0);
         sscanf(passIterThousands, "%lu", &iter);
         iter *= 1000;
         while(iter > 0) {
             /* encrypt both 128bit blocks with AES-256 */
-            aes_encrypt(&ctx, &hb[ 0], &hb[ 0]);
-            aes_encrypt(&ctx, &hb[16], &hb[16]);
+            aes_encrypt(&ctx, &hbu.hb[ 0], &hbu.hb[ 0]);
+            aes_encrypt(&ctx, &hbu.hb[16], &hbu.hb[16]);
             /* exchange upper half of first block with lower half of second block */
-            memcpy(&tempkey[0], &hb[8], 8);
-            memcpy(&hb[8], &hb[16], 8);
-            memcpy(&hb[16], &tempkey[0], 8);
+            memcpy(&tku.tempkey[0], &hbu.hb[8], 8);
+            memcpy(&hbu.hb[8], &hbu.hb[16], 8);
+            memcpy(&hbu.hb[16], &tku.tempkey[0], 8);
             iter--;
         }
-        memset(&tempkey[0], 0, sizeof(tempkey));
+        memset(&tku.tempkey[0], 0, sizeof(tku.tempkey));
     }
-    aes_set_key(&ctx, hb, bits, 0);
-    memset(hb, 0, sizeof(hb));
+    aes_set_key(&ctx, &hbu.hb[0], bits, 0);
+    memset(&hbu.hb[0], 0, sizeof(hbu.hb));
     memset(pass, 0, x);
 
     ret = 0;
@@ -694,5 +760,6 @@ int main(int argc, char **argv)
     memset(&multiKeyCtx[0], 0, sizeof(multiKeyCtx));
     memset(&iv[0], 0, sizeof(iv));
     memset(&buf.b[0], 0, sizeof(buf));
+    memset(&partialMD5[0], 0, sizeof(partialMD5));
     exit(ret);
 }
